/*
 * include file for pixel data
 */
#include "vort.h"

/*
 * types and defs for the ray tracer
 */

#define	TRUE	1
#define	FALSE	0

#define	MESLEN	100

/*
 * distance we must be from the last surface we hit before we register
 * another hit.
 */
#define	TOLERANCE	0.0001

/*
 * where we put infinity
 */
#define	HUGE_DIST	10000000.0

typedef float	matrix[4][4];
typedef float	mat3x3[3][3];

typedef struct v {
	float	x, y, z;
} vector;

typedef struct col {
	float	r, g, b;
} colour;

typedef struct pix {
	float	r, g, b, a;
} pixel;

/*
 * typedefs for texture, bumps, and tile patterns
 */
typedef struct t {
	int		type;
	char		axis;		/* major axis of normal to polygon */
	int		refset;		/* tile reference points set? */
	int		octaves;	/* range for noise function */
	mat3x3		mat;		/* rotation matrix to texture space */
	vector		trans, scales;	/* translation and scaling values */
	colour		c;
	colour		a;
	colour		blendcolour;
	float		var[3];
	vector		vref1, vref2;
	float		scalew, scaleh;
	char		*map;
	unsigned short	pixw, pixh;
} texture;

typedef struct tl {
	void		(*txtcol)();	/* texturing function */
	texture		txt;		/* texture */
	struct tl	*nxt;
} tlist;
	

/*
 * general structure for a surface
 */
typedef struct surf {
	colour	c, a;			/* surface and ambient light colour */
	float	kd, ks;			/* diffuse, specular */
	int	ksexp;			/* specular exponent */
	float	refl,			/* reflection */
		ri,			/* refractive index */
		absorb,			/* % of light absorbed in unit length */
		trans;			/* transparency */
	tlist	*txtlist;		/* texturing info */
} surface;

/*
 * attribute stack structure
 */
typedef struct at {
	matrix	m;		/* current transformation matrix */
	int	mcopy;		/* do we need to copy the matrix? */
	int	options;	/* shading options */
	float	maxscale;	/* maximum scale applied to an object */
	vector	scales,		/* scaling factors */
		trans;		/* translation factors */
	surface	s;		/* surface properties */
} attr;

/*
 * option flags
 */
#define	SHADOWS	1		/* for lights - shadow testing on/off */

/*
 * nodes for storing hit lists
 */

#define	ENTRY	1
#define	EXIT	2

typedef struct hl {
	float		t;
	int		type;
	struct o	*obj;
	struct hl	*nxt;
} hlist;

/*
 * defs for lights
 */
#define	POINT		0
#define	DIRECTIONAL	1
#define	DISTANT		2

typedef struct l {
	int		type;
	vector		org;
	vector		dir;
	colour		c;
	float		rad;
	float		cosang;
	int		rays;
	struct o	**lasthits;
	struct l	*nxt;
} light;

/*
 * the ray structure
 */
typedef struct r {
	short	raynumber;
	vector	org, dir;
} ray;

/*
 * major axis flags
 */
#define	XAXIS	0
#define	YAXIS	1
#define	ZAXIS	2

#define	DIMS	3

/*
 * hit types
 */

/*
 * for boxes
 */
#define	XFACE	0
#define	YFACE	1
#define	ZFACE	2
#define	NOFACE	3

/*
 * for quadrics
 */
#define	FACE	1
#define	SIDE	2

/*
 * object model definitions
 */

typedef struct s {
	float	rad, radsqu;
	vector	cent;
} sphere;

/*
 * typedefs for polygons and polygon models
 */
#define	MAXVERTS	256		/* maximum number of vertices */

typedef unsigned short	vertno;

typedef struct ed {
	vector		n;		/* normal */
	float		cnst;		/* plane constant */
	struct ed	*nxt;
} edge;

typedef struct fct {
	vector		n;		/* normal */
	float		cnst;		/* const for the face's plane */
	edge		*edges;		/* edge planes */
} polygon;

/*
 * details to go in objects
 */
typedef struct det {
	int		type;
	union {
		vector		v;
		float		f;
		char		*s;
		struct o	*obj;
	} u;
	struct det	*nxt;
} details;

/*
 * general object structure and defs
 */
typedef struct o {
	int		type;
	surface		s;
	short		raynumber;		/* number of last ray tested */
	float		lastt;
	hlist		*(*intersects)();
	void		(*normal)();
	union {
		sphere		*sph;
		polygon		*ply;
		struct bspt	*tree;
	} u;
	struct o	*nxt;
} object;

#ifndef M_PI
#define M_PI    3.14159265358979323846
#endif

#ifdef MSC

extern char	*smalloc();

#define yywrap()	1
#endif

/*
 * tree structure
 */
#define	X	0
#define	Y	1
#define	Z	2

#define DIMS	3

#define	LEAF	4
#define	SPLITABLELEAF	(0x08 | LEAF)

#define	MAXOBJS		3
#define MAXDEPTH	12

#define	splittable(a)	((a)->type & 0x08)

typedef struct b {
	float	min[DIMS];
	float	max[DIMS];
} bbox;

typedef struct ol {
	object		*obj;
	struct ol	*nxt;
} olist;

typedef struct bspt {
	int	type;
	float	splitval;
	bbox	bb;
	union {
		struct {
			int		depth;
			struct ol	*oblist;
		} leaf;
		struct {
			struct bspt	*left, *right;
		} branch;
	} u;
} stree;

